// Code generated by command: go run asm2.go -out search_amd64.s -stubs stub_search_amd64.go. DO NOT EDIT.

package simd

// Search finds the first idx for which xs[idx] >= k in xs.
func Search(xs []uint64, k uint64) int16
